/*
 * Avisynth script
 * MobileHackerz http://www.nurs.or.jp/~calcium/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "avformat.h"
#include <windows.h>
#include "avisynthdll.h"

#define AVS_ERRMSG_LEN 1024

typedef struct AVSContext {
    HMODULE        avsdll;            // DLL Handle
    void *(* avs_create)();
    int __stdcall (* avs_open)(void *,char *,AVSDLLVideoInfo *);
    int __stdcall (* avs_getvframe)(void *avs, int clip_num, int frm, AVSDLLVideoFrame *vf);
    int __stdcall (* avs_getaframe)(void *avs, int clip_num, void *buf, AVSDLLAudioPos *pos);
    int __stdcall (* avs_close)(void *avs, int clip_num);
    int __stdcall (* avs_getlasterror)(char *str,int len);

    void        *ctx;            // Script control class
    int         clip_num;        // Video clip number

    int         video_frame_num;
    int         max_frames;
    int         frame_raten;
    int         frame_rated;

    int         audio_frame_num;
    int64_t     audio_start;
    int64_t     num_audio_samples;

    int         video_stream;
    int         audio_stream;
    int         interleave_mode;

    char        errmsg[AVS_ERRMSG_LEN];
} AVSContext;

static void cut_crlf(char *str)
{
    int l = strlen(str),i,j = 0;
    for(i=0;i<=l;i++) {
        if ((str[i] != 13) && (str[i] != 10)) str[j++] = str[i];
    }
}

static int avs_probe(AVProbeData *p)
{
    char *ext;

    ext = (char *)p->filename; while(*ext != 0) ext++;
    ext -= 4;
    if (ext < p->filename) return 0;
    if ((ext[0] == '.') && ((ext[1]&0xdf) == 'A') && ((ext[2]&0xdf) == 'V') && ((ext[3]&0xdf) == 'S'))
        return AVPROBE_SCORE_MAX;

    return 0;
}

static int avs_read_header(AVFormatContext *s, AVFormatParameters *ap)
{
    int ret;
    AVSDLLVideoInfo vi;
    AVStream *st;
    AVSContext *avs = (AVSContext *) s->priv_data;

    // ----------------- Initialize
    avs->video_stream = -1;
    avs->audio_stream = -1;

    // ----------------- load DLL
    avs->avsdll                 = LoadLibrary("AVSredirect.dll");
    if (!avs->avsdll)           { av_log(s, AV_LOG_ERROR, "failed to load AVSredirect.dll\n"); return -1; }
    avs->avs_create             = (void *(*)()) GetProcAddress(avs->avsdll, "_avs_create@0");
    if (!avs->avs_create)       { av_log(s, AV_LOG_ERROR, "failed to load avs_create\n"); return -1; }
    avs->avs_open               = (int __stdcall (*)(void *,char *,AVSDLLVideoInfo *)) GetProcAddress(avs->avsdll, "_avs_open@12");
    if (!avs->avs_open)         { av_log(s, AV_LOG_ERROR, "failed to load avs_open\n"); return -1; }
    avs->avs_getvframe          = (int __stdcall (*)(void *,int,int,AVSDLLVideoFrame *)) GetProcAddress(avs->avsdll, "_avs_getvframe@16");
    if (!avs->avs_getvframe)    { av_log(s, AV_LOG_ERROR, "failed to load avs_getvframe\n"); return -1; }
    avs->avs_getaframe          = (int __stdcall (*)(void *, int, void *, AVSDLLAudioPos *)) GetProcAddress(avs->avsdll, "_avs_getaframe@16");
    if (!avs->avs_getaframe)    { av_log(s, AV_LOG_ERROR, "failed to load avs_getaframe\n"); return -1; }
    avs->avs_close              = (int __stdcall (*)(void *, int)) GetProcAddress(avs->avsdll, "_avs_close@8");
    if (!avs->avs_close)        { av_log(s, AV_LOG_ERROR, "failed to load avs_close\n"); return -1; }
    avs->avs_getlasterror       = (int __stdcall (*)(char *, int)) GetProcAddress(avs->avsdll, "_avs_getlasterror@8");
    if (!avs->avs_getlasterror) { av_log(s, AV_LOG_ERROR, "failed to load avs_getlasterror\n"); return -1; }

    // ----------------- open
    avs->ctx = avs->avs_create();
    if (!avs->ctx) {
        avs->avs_getlasterror(avs->errmsg,AVS_ERRMSG_LEN);
        cut_crlf(avs->errmsg);
        av_log(s, AV_LOG_ERROR, "failed to load Avisynth : %s\n",avs->errmsg);
        return -1;
    }

    vi.width = 1;
    ret = avs->avs_open(avs->ctx,s->filename,&vi);
    if (ret < 0) {
        avs->avs_getlasterror(avs->errmsg,AVS_ERRMSG_LEN);
        cut_crlf(avs->errmsg);
        av_log(s, AV_LOG_ERROR, "failed to open avs : %s\n",avs->errmsg);
        return -1;
    }
    avs->clip_num = ret;

    // ----------------- 
    if (vi.raten == 0 && vi.rated == 0) {
        // Frame rate unknown
        vi.raten = 30;
        vi.rated = 1;
    }

    if (vi.aspectn == 0 && vi.aspectd == 0) {
        // Pixel aspect unknown
        vi.aspectd = 1;
    }

    // ---------- Stream
    avs->video_frame_num   = 0;
    avs->audio_frame_num   = 0;
    avs->frame_raten       = vi.raten;
    avs->frame_rated       = vi.rated;
    avs->max_frames        = vi.num_frames;
    avs->num_audio_samples = vi.num_audio_samples;

    // ---------- Video
    if ((vi.width > 0) && (vi.height > 0)) {
        st = av_new_stream(s, 0);
        st->codec->width               = vi.width;
        st->codec->height              = vi.height;
        av_reduce(&vi.raten, &vi.rated, vi.raten, vi.rated, (1UL<<31)-1);
        st->codec->time_base.den       = vi.raten;
        st->codec->time_base.num       = vi.rated;
        st->codec->pix_fmt             = PIX_FMT_YUV420P;
        st->codec->codec_type          = CODEC_TYPE_VIDEO;
        st->codec->codec_id            = CODEC_ID_RAWVIDEO;
        st->codec->sample_aspect_ratio = (AVRational){vi.aspectn, vi.aspectd};

        st->r_frame_rate.den          = vi.raten;
        st->r_frame_rate.num          = vi.rated;
        av_set_pts_info(st, 64, vi.rated, vi.raten);
        st->start_time                = 0;
        st->duration                  = av_rescale(vi.num_frames, (int64_t)st->codec->time_base.num * AV_TIME_BASE, st->codec->time_base.den);

        avs->video_stream = st->index;
    }

    // ---------- Audio
    if (vi.audio_samples_per_second != 0) {
        st = av_new_stream(s, 1);
        st->codec->width               = 0;
        st->codec->height              = 0;
        st->codec->time_base.den       = 1;
        st->codec->time_base.num       = vi.audio_samples_per_second;
        st->codec->codec_type          = CODEC_TYPE_AUDIO;
        st->codec->codec_id            = CODEC_ID_PCM_S16LE;        // CODEC_ID_PCM_S16LE CODEC_ID_PCM_S16BE CODEC_ID_PCM_U16LE CODEC_ID_PCM_U16BE
        st->codec->sample_rate         = vi.audio_samples_per_second;
        st->codec->channels            = vi.nchannels;

        st->r_frame_rate.den          = 1;
        st->r_frame_rate.num          = vi.audio_samples_per_second;
        av_set_pts_info(st, 64, 1, st->codec->sample_rate);
        st->start_time                = 0;
        st->duration                  = av_rescale(vi.num_audio_samples, (int64_t)st->codec->time_base.den * AV_TIME_BASE, st->codec->time_base.num);

        avs->audio_start              = 0;
        avs->audio_stream             = st->index;
    }


    if ((avs->video_stream == -1) && (avs->audio_stream == -1)) {
        return -1;
    } else {
        if (avs->video_stream != -1) {
            avs->interleave_mode = 0;
        } else {
            avs->interleave_mode = 1;
        }
        return 0;
    }
}


static int avs_read_packet(AVFormatContext *s, AVPacket *pkt)
{
    AVSDLLVideoFrame vf;
    int packet_size, ret, width, height, p, y, wrote, w, i, retry;
    AVSContext *avs = (AVSContext *) s->priv_data;
    AVStream *st;
    char *buf,*data;
    int processed = 0;
    AVSDLLAudioPos pos;

    for(retry=0;(retry<2)&&(processed==0);retry++) {
        // Video
        if ((avs->interleave_mode == 0) && (processed == 0)) {
            if (avs->video_stream != -1) {
                st = s->streams[avs->video_stream];
                width = st->codec->width;
                height = st->codec->height;
        
                if (avs->video_frame_num < avs->max_frames) {
                    packet_size = avpicture_get_size(st->codec->pix_fmt, width, height);
                    if (packet_size < 0)
                        return -1;
        
                    if (av_new_packet(pkt, packet_size) < 0)
                        return AVERROR_IO;
        
                    ret = avs->avs_getvframe(avs->ctx,avs->clip_num,avs->video_frame_num,&vf);
                    if (ret != 0) {
                        avs->avs_getlasterror(avs->errmsg,AVS_ERRMSG_LEN);
                        cut_crlf(avs->errmsg);
                        av_log(s, AV_LOG_ERROR, "avs_getvframe failed : %s\n",avs->errmsg);
                        av_free_packet(pkt);
                        return AVERROR_IO;
                    }
        
                    pkt->stream_index = avs->video_stream;
                    pkt->duration     = av_rescale(avs->video_frame_num+1, (int64_t)avs->frame_rated * AV_TIME_BASE, avs->frame_raten)
                                      - av_rescale(avs->video_frame_num, (int64_t)avs->frame_rated * AV_TIME_BASE, avs->frame_raten);
                    pkt->pts          = av_rescale(avs->video_frame_num, (int64_t)avs->frame_rated * AV_TIME_BASE, avs->frame_raten);
                    pkt->dts          = pkt->pts;
        
                    buf = pkt->data;
                    for(p=0;p<3;p++) {
                        w     = vf.plane[p].width;
                        data  = vf.plane[p].data;
                        wrote = 0;
                        for(y=0; y<vf.plane[p].height; y++) {
                            if ((wrote+w)<=pkt->size) {
                                for(i=0;i<w;i++) buf[i] = data[i];
                            }
                            buf   += w;
                            data  += vf.plane[p].pitch;
                            wrote += w;
                        }
                    }
    
                    avs->video_frame_num++;
                    processed = 1;
                }
            }
            if (avs->audio_stream != -1) avs->interleave_mode = 1;
        }

        // Audio
        if ((avs->interleave_mode == 1) && (processed == 0)) {
            if (avs->audio_stream != -1) {
                st = s->streams[avs->audio_stream];
                pos.start = av_rescale(avs->audio_frame_num, (int64_t)avs->frame_rated * st->codec->time_base.num, avs->frame_raten);
                pos.count = av_rescale(avs->audio_frame_num+1, (int64_t)avs->frame_rated * st->codec->time_base.num, avs->frame_raten) - pos.start;
    
                if ((pos.start+pos.count) > avs->num_audio_samples) {
                    pos.count = avs->num_audio_samples - pos.start;
                }
    
                if (pos.count > 0) {
                    packet_size = pos.count * 2 * (int64_t)st->codec->channels;        // 16bit
                    if (av_new_packet(pkt, packet_size) < 0)
                        return AVERROR_IO;
    
                    ret = avs->avs_getaframe(avs->ctx,avs->clip_num,pkt->data,&pos);
                    if (ret != 0) {
                        avs->avs_getlasterror(avs->errmsg,AVS_ERRMSG_LEN);
                        cut_crlf(avs->errmsg);
                        av_log(s, AV_LOG_ERROR, "avs_getaframe failed : %s\n",avs->errmsg);
                        av_free_packet(pkt);
                        return AVERROR_IO;
                    }
    
                    pkt->stream_index = avs->audio_stream;
                    pkt->duration     = av_rescale(avs->audio_frame_num+1, (int64_t)avs->frame_rated * AV_TIME_BASE, avs->frame_raten)
                                      - av_rescale(avs->audio_frame_num, (int64_t)avs->frame_rated * AV_TIME_BASE, avs->frame_raten);
                    pkt->pts          = av_rescale(avs->audio_frame_num, (int64_t)avs->frame_rated * AV_TIME_BASE, avs->frame_raten);
                    pkt->dts          = pkt->pts;
    
                    avs->audio_frame_num++;
                    processed = 1;
                }
            }
            if (avs->video_stream != -1) avs->interleave_mode = 0;
        }
    }

    // Fin
    if (processed != 0) {
        return 0;
    } else {
        return -1;
    }
}

static int avs_read_close(AVFormatContext *s)
{
    AVSContext *avs = (AVSContext *) s->priv_data;

    avs->avs_close(avs->ctx,avs->clip_num);
//    if (avs->avsdll) FreeLibrary(avs->avsdll);
    return 0;
}

static int avs_read_seek(AVFormatContext *s, 
                  int stream_index, int64_t timestamp, int flags)
{
    AVStream *st;
    AVSContext *avs = (AVSContext *) s->priv_data;
    int    frame;
    int64_t ts;

    st = s->streams[stream_index];
    ts = av_rescale(timestamp, (int64_t)st->time_base.num * AV_TIME_BASE, st->time_base.den);
    frame = av_rescale(ts, avs->frame_raten, (int64_t)avs->frame_rated * AV_TIME_BASE);

    if (avs->video_stream != -1) {
        avs->video_frame_num = frame;
        s->streams[avs->video_stream]->start_time = ts;
    }
    if (avs->audio_stream != -1) {
        avs->audio_frame_num = frame;
        s->streams[avs->audio_stream]->start_time = ts;
    }
    return 0;
}

AVInputFormat avs_iformat = {
    "avs",
    "Avisynth script",
    sizeof(AVSContext),
    avs_probe,                // probe
    avs_read_header,
    avs_read_packet,
    avs_read_close,
    avs_read_seek,
    .extensions = "avs"
};

int avs_init(void)
{
    av_register_input_format(&avs_iformat);
    return 0;
}

